import re
from django.apps import apps
from django.conf import settings


def get_model(const, default):
    model_str = getattr(settings, const, default)
    if not re.match("\w+\.\w+", model_str):
        raise LookupError("%s setting must be of the form 'app.Model'" % (const))
    return apps.get_model(*model_str.split("."))


def get_address_model():
    return get_model("ADDRESS_MODEL", "uk_geo_utils.Address")


def get_onsud_model():
    return get_model("ONSUD_MODEL", "uk_geo_utils.Onsud")


def get_onspd_model():
    return get_model("ONSPD_MODEL", "uk_geo_utils.Onspd")


class Postcode:
    def __init__(self, postcode, validate=False):
        self.postcode = re.sub("[^A-Z0-9]", "", str(postcode).upper())
        if validate and len(str(self.postcode)) < 5:
            raise ValueError("Postcode must have at least 5 characters")

    def __str__(self):
        return self.without_space

    def __eq__(self, other):
        return (
            type(self) == Postcode
            and type(other) == Postcode
            and self.without_space == other.without_space
        )

    @property
    def territory(self):
        if self.postcode[:2] == "BT":
            return "NI"
        return "GB"

    @property
    def with_space(self):
        return self.postcode[:-3] + " " + self.postcode[-3:]

    @property
    def without_space(self):
        return self.postcode


class AddressFormatter:
    def __init__(
        self,
        organisation_name,
        department_name,
        po_box_number,
        sub_building_name,
        building_name,
        building_number,
        dependent_thoroughfare,
        thoroughfare,
        post_town,
        double_dependent_locality,
        dependent_locality,
    ):
        """one to one mapping."""
        self.organisation_name = organisation_name
        self.department_name = department_name
        self.po_box_number = po_box_number
        self.sub_building_name = sub_building_name
        self.building_name = building_name
        self.building_number = building_number
        self.dependent_thoroughfare = dependent_thoroughfare
        self.thoroughfare = thoroughfare
        self.post_town = post_town
        self.double_dependent_locality = double_dependent_locality
        self.dependent_locality = dependent_locality
        self.address_label = []

    def generate_address_label(self):
        """Construct a list for address label.

        Non-empty premises elements are appended to the address label in the
        order of organisation_name, department_name, po_box_number (which
        must be prepended with 'PO Box', sub_building_name, building_name,
        building_number, then the rest of the elements except for Town and
        Postcode because we want them in their own fields. This isn't strict
        address label but we're probably loading them into a database.
        """
        if self.organisation_name:
            self.address_label.append(self.organisation_name)
        if self.department_name:
            self.address_label.append(self.department_name)
        if self.po_box_number:
            self.address_label.append("PO Box " + self.po_box_number)

        elements = [
            self.sub_building_name,
            self.building_name,
            self.building_number,
            self.dependent_thoroughfare,
            self.thoroughfare,
            self.double_dependent_locality,
            self.dependent_locality,
        ]

        for element in elements:
            if element:
                self._append_to_label(element)

        # pad label to length of 7 if not already
        if len(self.address_label) < 7:
            for i in range(7 - len(self.address_label)):
                self.address_label.append("")

        # finally, add post town
        self.address_label[5] = self.post_town

        return ", ".join([f for f in self.address_label if f])

    def _is_exception_rule(self, element):
        """ Check for "exception rule".

        Address elements will be appended onto a new line on the lable except
        for when the penultimate lable line fulfils certain criteria, in which
        case the element will be concatenated onto the penultimate line. This
        method checks for those criteria.

        i) First and last characters of the Building Name are numeric
          (eg '1to1' or '100:1')
        ii) First and penultimate characters are numeric, last character is
          alphabetic (eg '12A')
        iii) Building Name has only one character (eg 'A')
        """
        if element[0].isdigit() and element[-1].isdigit():
            return True
        if (
            len(element) > 1
            and element[0].isdigit()
            and element[-2].isdigit()
            and element[-1].isalpha()
        ):
            return True
        if len(element) == 1 and element.isalpha():
            return True
        return False

    def _append_to_label(self, element):
        """Append address element to the label.

        Normally an element will be appended onto the list, except where the
        existing last element fulfils the exception rule, in which case the
        element will be concatenated onto the final list member.
        """
        if len(self.address_label) > 0 and self._is_exception_rule(
            self.address_label[-1]
        ):
            self.address_label[-1] += " " + element
        else:
            self.address_label.append(element)

    def __str__(self):
        """Return the label form of the address."""
        return ",".join(self.generate_address_label())


class AddressSorter:
    # Class for sorting sort a list of address objects
    # in a human-readable order.

    def __init__(self, addresses):
        self.addresses = addresses

    def convert(self, text):
        # if text is numeric, covert to an int
        # this allows us to sort numbers in int order, not string order
        return int(text) if text.isdigit() else text

    def alphanum_key(self, tup):
        # split the desired component of tup (defined by key function)
        # into a listof numeric and text components
        return [self.convert(c) for c in filter(None, re.split("([0-9]+)", tup[1]))]

    def swap_fields(self, item):
        lst = self.alphanum_key(item)
        # swap things about so we can sort by street name, house number
        # instead of house number, street name
        if (
            len(lst) > 1
            and isinstance(lst[0], int)
            and isinstance(lst[1], str)
            and (lst[1][0].isspace() or lst[1][0] == ",")
        ):
            lst[0], lst[1] = lst[1], lst[0]
        if len(lst) > 1 and isinstance(lst[0], int) and isinstance(lst[1], int):
            lst[0], lst[1] = lst[1], lst[0]
        if isinstance(lst[0], int):
            lst[0] = str(lst[0])
        return lst

    def natural_sort(self):
        sorted_list = sorted(
            [(address, address.address) for address in self.addresses],
            key=self.swap_fields,
        )
        return [address[0] for address in sorted_list]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.contrib.gis.db.models.fields


class Migration(migrations.Migration):

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Address",
            fields=[
                (
                    "uprn",
                    models.CharField(max_length=100, serialize=False, primary_key=True),
                ),
                ("address", models.TextField(blank=True)),
                (
                    "postcode",
                    models.CharField(max_length=15, db_index=True, blank=True),
                ),
                (
                    "location",
                    django.contrib.gis.db.models.fields.PointField(
                        null=True, srid=4326, blank=True
                    ),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="Onsud",
            fields=[
                (
                    "uprn",
                    models.CharField(max_length=12, serialize=False, primary_key=True),
                ),
                ("ctry_flag", models.CharField(max_length=1, blank=True)),
                ("cty", models.CharField(max_length=9, blank=True)),
                ("lad", models.CharField(max_length=9, blank=True)),
                ("ward", models.CharField(max_length=9, blank=True)),
                ("hlthau", models.CharField(max_length=9, blank=True)),
                ("ctry", models.CharField(max_length=9, blank=True)),
                ("rgn", models.CharField(max_length=9, blank=True)),
                ("pcon", models.CharField(max_length=9, blank=True)),
                ("eer", models.CharField(max_length=9, blank=True)),
                ("ttwa", models.CharField(max_length=9, blank=True)),
                ("nuts", models.CharField(max_length=9, blank=True)),
                ("park", models.CharField(max_length=9, blank=True)),
                ("oa11", models.CharField(max_length=9, blank=True)),
                ("lsoa11", models.CharField(max_length=9, blank=True)),
                ("msoa11", models.CharField(max_length=9, blank=True)),
                ("parish", models.CharField(max_length=9, blank=True)),
                ("wz11", models.CharField(max_length=9, blank=True)),
                ("ccg", models.CharField(max_length=9, blank=True)),
                ("bua11", models.CharField(max_length=9, blank=True)),
                ("buasd11", models.CharField(max_length=9, blank=True)),
                ("ruc11", models.CharField(max_length=2, blank=True)),
                ("oac11", models.CharField(max_length=3, blank=True)),
                ("lep1", models.CharField(max_length=9, blank=True)),
                ("lep2", models.CharField(max_length=9, blank=True)),
                ("pfa", models.CharField(max_length=9, blank=True)),
                ("imd", models.CharField(max_length=5, blank=True)),
            ],
            options={"abstract": False},
        ),
    ]



# Generated by Django 2.0.8 on 2019-02-15 11:21

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("uk_geo_utils", "0006_auto_20180612_1655")]

    operations = [migrations.RemoveField(model_name="onsud", name="ctry_flag")]



import os
from django.db import connection
from django.db import transaction
from django.core.management.base import BaseCommand
from uk_geo_utils.helpers import get_address_model


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "cleaned_ab_path",
            help="The path to the folder containing the cleaned AddressBase CSVs",
        )
        parser.add_argument(
            "-t",
            "--transaction",
            help="Run the import in a transaction",
            action="store_true",
            default=False,
            dest="transaction",
        )

    def handle(self, *args, **kwargs):
        self.table_name = get_address_model()._meta.db_table
        self.path = kwargs["cleaned_ab_path"]
        if kwargs["transaction"]:
            with transaction.atomic():
                self.import_addressbase()
        else:
            self.import_addressbase()

    def import_addressbase(self):
        cleaned_file_path = os.path.abspath(
            os.path.join(self.path, "addressbase_cleaned.csv")
        )

        with open(cleaned_file_path, "r") as fp:

            cursor = connection.cursor()
            self.stdout.write("clearing existing data..")
            cursor.execute("TRUNCATE TABLE %s;" % (self.table_name))

            self.stdout.write("importing from %s.." % (cleaned_file_path))
            cursor.copy_expert(
                """
                COPY %s (UPRN,address,postcode,location)
                FROM STDIN (FORMAT CSV, DELIMITER ',', quote '"');
            """
                % (self.table_name),
                fp,
            )

        self.stdout.write("...done")



import os
import glob
from django.db import connection
from django.db import transaction
from django.core.management.base import BaseCommand
from uk_geo_utils.helpers import get_onsud_model


class Command(BaseCommand):
    """
    To import ONSUD, grab the latest release:
    http://ons.maps.arcgis.com/home/search.html?q=ONS%20Address%20Directory&t=content
    and run
    python manage.py import_onsud /path/to/data
    """

    def add_arguments(self, parser):
        parser.add_argument(
            "path", help="Path to the directory containing the ONSUD CSVs"
        )
        parser.add_argument(
            "-t",
            "--transaction",
            help="Run the import in a transaction",
            action="store_true",
            default=False,
            dest="transaction",
        )

    def handle(self, *args, **kwargs):
        self.table_name = get_onsud_model()._meta.db_table
        self.path = kwargs["path"]
        if kwargs["transaction"]:
            with transaction.atomic():
                self.import_onsud()
        else:
            self.import_onsud()

    def import_onsud(self):
        self.table_name = get_onsud_model()._meta.db_table

        glob_str = os.path.join(self.path, "*.csv")
        files = glob.glob(glob_str)
        if not files:
            raise FileNotFoundError("No CSV files found in %s" % (self.path))

        cursor = connection.cursor()
        self.stdout.write("clearing existing data..")
        cursor.execute("TRUNCATE TABLE %s;" % (self.table_name))

        self.stdout.write("importing from files..")
        for f in files:
            self.stdout.write(f)
            with open(f, "r") as fp:
                cursor.copy_expert(
                    """
                    COPY %s (
                    uprn, cty, ced, lad, ward, parish, hlthau, ctry,
                    rgn, pcon, eer, ttwa, nuts, park, oa11, lsoa11, msoa11,
                    wz11, ccg, bua11, buasd11, ruc11, oac11, lep1, lep2, pfa, imd)
                    FROM STDIN (FORMAT CSV, DELIMITER ',', QUOTE '"', HEADER);
                """
                    % (self.table_name),
                    fp,
                )
        self.stdout.write("...done")



# -*- coding: utf-8 -*-
# Generated by Django 1.11.10 on 2018-03-05 16:33
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("uk_geo_utils", "0004_auto_20180228_1553")]

    operations = [
        migrations.AddField(
            model_name="onsud",
            name="ced",
            field=models.CharField(blank=True, max_length=9),
        )
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("uk_geo_utils", "0002_onspd")]

    operations = [
        migrations.RemoveField(model_name="onspd", name="id"),
        migrations.AlterField(
            model_name="onspd",
            name="pcds",
            field=models.CharField(
                primary_key=True, max_length=8, blank=True, serialize=False
            ),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.contrib.gis.db.models.fields


class Migration(migrations.Migration):

    dependencies = [("uk_geo_utils", "0001_initial")]

    operations = [
        migrations.CreateModel(
            name="Onspd",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        verbose_name="ID",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("pcd", models.CharField(blank=True, max_length=7)),
                ("pcd2", models.CharField(blank=True, max_length=8)),
                ("pcds", models.CharField(blank=True, max_length=8)),
                ("dointr", models.CharField(blank=True, max_length=6)),
                ("doterm", models.CharField(blank=True, max_length=6)),
                ("oscty", models.CharField(blank=True, max_length=9)),
                ("oslaua", models.CharField(blank=True, max_length=9)),
                ("osward", models.CharField(blank=True, max_length=9)),
                ("usertype", models.CharField(blank=True, max_length=1)),
                ("oseast1m", models.CharField(blank=True, max_length=6)),
                ("osnrth1m", models.CharField(blank=True, max_length=7)),
                ("osgrdind", models.CharField(blank=True, max_length=1)),
                ("oshlthau", models.CharField(blank=True, max_length=9)),
                ("hro", models.CharField(blank=True, max_length=9)),
                ("ctry", models.CharField(blank=True, max_length=9)),
                ("gor", models.CharField(blank=True, max_length=9)),
                ("streg", models.CharField(blank=True, max_length=1)),
                ("pcon", models.CharField(blank=True, max_length=9)),
                ("eer", models.CharField(blank=True, max_length=9)),
                ("teclec", models.CharField(blank=True, max_length=9)),
                ("ttwa", models.CharField(blank=True, max_length=9)),
                ("pct", models.CharField(blank=True, max_length=9)),
                ("nuts", models.CharField(blank=True, max_length=10)),
                ("psed", models.CharField(blank=True, max_length=8)),
                ("cened", models.CharField(blank=True, max_length=6)),
                ("edind", models.CharField(blank=True, max_length=1)),
                ("oshaprev", models.CharField(blank=True, max_length=3)),
                ("lea", models.CharField(blank=True, max_length=3)),
                ("oldha", models.CharField(blank=True, max_length=3)),
                ("wardc91", models.CharField(blank=True, max_length=6)),
                ("wardo91", models.CharField(blank=True, max_length=6)),
                ("ward98", models.CharField(blank=True, max_length=6)),
                ("statsward", models.CharField(blank=True, max_length=6)),
                ("oa01", models.CharField(blank=True, max_length=10)),
                ("casward", models.CharField(blank=True, max_length=6)),
                ("park", models.CharField(blank=True, max_length=9)),
                ("lsoa01", models.CharField(blank=True, max_length=9)),
                ("msoa01", models.CharField(blank=True, max_length=9)),
                ("ur01ind", models.CharField(blank=True, max_length=1)),
                ("oac01", models.CharField(blank=True, max_length=3)),
                ("oldpct", models.CharField(blank=True, max_length=5)),
                ("oa11", models.CharField(blank=True, max_length=9)),
                ("lsoa11", models.CharField(blank=True, max_length=9)),
                ("msoa11", models.CharField(blank=True, max_length=9)),
                ("parish", models.CharField(blank=True, max_length=9)),
                ("wz11", models.CharField(blank=True, max_length=9)),
                ("ccg", models.CharField(blank=True, max_length=9)),
                ("bua11", models.CharField(blank=True, max_length=9)),
                ("buasd11", models.CharField(blank=True, max_length=9)),
                ("ru11ind", models.CharField(blank=True, max_length=2)),
                ("oac11", models.CharField(blank=True, max_length=3)),
                ("lat", models.CharField(blank=True, max_length=10)),
                ("long", models.CharField(blank=True, max_length=10)),
                ("lep1", models.CharField(blank=True, max_length=9)),
                ("lep2", models.CharField(blank=True, max_length=9)),
                ("pfa", models.CharField(blank=True, max_length=9)),
                ("imd", models.CharField(blank=True, max_length=5)),
                (
                    "location",
                    django.contrib.gis.db.models.fields.PointField(
                        null=True, blank=True, srid=4326
                    ),
                ),
            ],
            options={"abstract": False},
        )
    ]



from django.contrib.gis.db import models

try:
    from django.contrib.gis.db.models.manager import GeoManager
except ImportError:
    from django.db.models import Manager as GeoManager


class CachedGetMixin:
    def get_cached(self, pk):
        for record in self:
            if record.pk == pk:
                return record
        raise self.model.DoesNotExist()


class AddressQuerySet(models.QuerySet, CachedGetMixin):
    @property
    def centroid(self):
        if not self:
            return None

        if len(self) == 1:
            return self[0].location

        base_point = self[0].location
        poly = base_point.union(self[1].location)
        for m in self:
            poly = poly.union(m.location)

        return poly.centroid


class AbstractAddressManager(GeoManager):
    def get_queryset(self):
        return AddressQuerySet(self.model, using=self._db)


class AbstractAddress(models.Model):
    uprn = models.CharField(primary_key=True, max_length=100)
    address = models.TextField(blank=True)
    postcode = models.CharField(blank=True, max_length=15, db_index=True)
    location = models.PointField(null=True, blank=True)
    objects = AbstractAddressManager()

    class Meta:
        abstract = True


class Address(AbstractAddress):
    pass


class OnsudQuerySet(models.QuerySet, CachedGetMixin):
    pass


class AbstractOnsudManager(GeoManager):
    def get_queryset(self):
        return OnsudQuerySet(self.model, using=self._db)


class AbstractOnsud(models.Model):
    uprn = models.CharField(primary_key=True, max_length=12)
    cty = models.CharField(blank=True, max_length=9)
    ced = models.CharField(blank=True, max_length=9)
    lad = models.CharField(blank=True, max_length=9)
    ward = models.CharField(blank=True, max_length=9)
    parish = models.CharField(blank=True, max_length=9)
    hlthau = models.CharField(blank=True, max_length=9)
    ctry = models.CharField(blank=True, max_length=9)
    rgn = models.CharField(blank=True, max_length=9)
    pcon = models.CharField(blank=True, max_length=9)
    eer = models.CharField(blank=True, max_length=9)
    ttwa = models.CharField(blank=True, max_length=9)
    nuts = models.CharField(blank=True, max_length=9)
    park = models.CharField(blank=True, max_length=9)
    oa11 = models.CharField(blank=True, max_length=9)
    lsoa11 = models.CharField(blank=True, max_length=9)
    msoa11 = models.CharField(blank=True, max_length=9)
    wz11 = models.CharField(blank=True, max_length=9)
    ccg = models.CharField(blank=True, max_length=9)
    bua11 = models.CharField(blank=True, max_length=9)
    buasd11 = models.CharField(blank=True, max_length=9)
    ruc11 = models.CharField(blank=True, max_length=2)
    oac11 = models.CharField(blank=True, max_length=3)
    lep1 = models.CharField(blank=True, max_length=9)
    lep2 = models.CharField(blank=True, max_length=9)
    pfa = models.CharField(blank=True, max_length=9)
    imd = models.CharField(blank=True, max_length=5)
    objects = AbstractOnsudManager()

    def _get_oscty(self):
        return self.cty

    def _get_oslaua(self):
        return self.lad

    def _get_osward(self):
        return self.ward

    def _get_oshlthau(self):
        return self.hlthau

    def _get_ru11ind(self):
        return self.ruc11

    oscty = property(_get_oscty)
    oslaua = property(_get_oslaua)
    osward = property(_get_osward)
    oshlthau = property(_get_oshlthau)
    ru11ind = property(_get_ru11ind)

    class Meta:
        abstract = True


class Onsud(AbstractOnsud):
    pass


class AbstractOnspd(models.Model):
    pcd = models.CharField(blank=True, max_length=7)
    pcd2 = models.CharField(blank=True, max_length=8)
    pcds = models.CharField(blank=True, max_length=8, primary_key=True)
    dointr = models.CharField(blank=True, max_length=6)
    doterm = models.CharField(blank=True, max_length=6)
    oscty = models.CharField(blank=True, max_length=9)
    ced = models.CharField(blank=True, max_length=9)
    oslaua = models.CharField(blank=True, max_length=9)
    osward = models.CharField(blank=True, max_length=9)
    parish = models.CharField(blank=True, max_length=9)
    usertype = models.CharField(blank=True, max_length=1)
    oseast1m = models.CharField(blank=True, max_length=6)
    osnrth1m = models.CharField(blank=True, max_length=7)
    osgrdind = models.CharField(blank=True, max_length=1)
    oshlthau = models.CharField(blank=True, max_length=9)
    nhser = models.CharField(blank=True, max_length=9)
    ctry = models.CharField(blank=True, max_length=9)
    rgn = models.CharField(blank=True, max_length=9)
    streg = models.CharField(blank=True, max_length=1)
    pcon = models.CharField(blank=True, max_length=9)
    eer = models.CharField(blank=True, max_length=9)
    teclec = models.CharField(blank=True, max_length=9)
    ttwa = models.CharField(blank=True, max_length=9)
    pct = models.CharField(blank=True, max_length=9)
    nuts = models.CharField(blank=True, max_length=10)
    statsward = models.CharField(blank=True, max_length=6)
    oa01 = models.CharField(blank=True, max_length=10)
    casward = models.CharField(blank=True, max_length=6)
    park = models.CharField(blank=True, max_length=9)
    lsoa01 = models.CharField(blank=True, max_length=9)
    msoa01 = models.CharField(blank=True, max_length=9)
    ur01ind = models.CharField(blank=True, max_length=1)
    oac01 = models.CharField(blank=True, max_length=3)
    oa11 = models.CharField(blank=True, max_length=9)
    lsoa11 = models.CharField(blank=True, max_length=9)
    msoa11 = models.CharField(blank=True, max_length=9)
    wz11 = models.CharField(blank=True, max_length=9)
    ccg = models.CharField(blank=True, max_length=9)
    bua11 = models.CharField(blank=True, max_length=9)
    buasd11 = models.CharField(blank=True, max_length=9)
    ru11ind = models.CharField(blank=True, max_length=2)
    oac11 = models.CharField(blank=True, max_length=3)
    lat = models.CharField(blank=True, max_length=10)
    long = models.CharField(blank=True, max_length=10)
    lep1 = models.CharField(blank=True, max_length=9)
    lep2 = models.CharField(blank=True, max_length=9)
    pfa = models.CharField(blank=True, max_length=9)
    imd = models.CharField(blank=True, max_length=5)
    calncv = models.CharField(blank=True, max_length=9)
    stp = models.CharField(blank=True, max_length=9)
    location = models.PointField(null=True, blank=True)
    objects = GeoManager()

    def _get_cty(self):
        """
        Note: these are not _exactly_ the same because

        'CTY' in ONSUD is:
        E10000002 - E10000034 = England (county);
        E11000001 - E11000007 = England (metropolitan county);
        E13000001 - E13000002 = England (Inner/Outer London);
        E99999999 (pseudo) = England (UA/MD/LB);
        W99999999 (pseudo) = Wales;
        S99999999 (pseudo) = Scotland

        whereas 'OSCTY' in ONSPD is
        E10000002 – E10000034 = England;
        E99999999 (pseudo) = England (UA/MD/LB);
        W99999999 (pseudo) = Wales;
        S99999999 (pseudo) = Scotland;
        """
        return self.oscty

    def _get_lad(self):
        return self.oslaua

    def _get_ward(self):
        return self.osward

    def _get_hlthau(self):
        return self.oshlthau

    def _get_ruc11(self):
        return self.ru11ind

    cty = property(_get_cty)
    lad = property(_get_lad)
    ward = property(_get_ward)
    hlthau = property(_get_hlthau)
    ruc11 = property(_get_ruc11)

    class Meta:
        abstract = True


class Onspd(AbstractOnspd):
    pass
